import hashlib
import os
import shutil
from typing import List


def hex_digest_files(filenames: List[str], algorithm_name: str) -> str:
    block_size = 65536
    hash_object = hashlib.new(algorithm_name)
    for filename in filenames:
        with open(filename, 'rb') as file_handle:
            buf = file_handle.read(block_size)
            while len(buf) > 0:
                hash_object.update(buf)
                buf = file_handle.read(block_size)
    return hash_object.hexdigest()


def hex_digest_file(filename: str, algorithm_name: str) -> str:
    return hex_digest_files(filenames=[filename], algorithm_name=algorithm_name)


def sha1_file(filename: str) -> str:
    return hex_digest_file(filename=filename, algorithm_name="sha1")


def sha1_files(filenames: List[str]) -> str:
    return hex_digest_files(filenames=filenames, algorithm_name="sha1")


def makedirs_for_file(filename: str):
    folder = os.path.dirname(filename)
    if folder != "":
        os.makedirs(folder, exist_ok=True)


def copy_mkdir(source: str, destination: str):
    makedirs_for_file(destination)
    shutil.copy(source, destination)


def unlink_files(files: List[str], only_if_exist: bool=True) -> None:
    for file in files:
        if only_if_exist:
            if os.path.isfile(file):
                os.unlink(file)
        else:
            os.unlink(file)


def files_under_folder(folder: str) -> List[str]:
    f = []
    for (dir_path, _dir_names, filenames) in os.walk(folder):
        f.extend([os.path.join(dir_path, filename) for filename in filenames])
    return f


def files_under_folders(folders: List[str]) -> List[str]:
    f = []
    for folder in folders:
        f.extend(files_under_folder(folder))
    return f


def sha1_folders(folders: List[str]) -> str:
    return sha1_files(filenames=files_under_folders(folders=folders))


def sha1_files_folders(files: List[str], folders: List[str]) -> str:
    f = files
    f.extend(files_under_folders(folders=folders))
    return sha1_files(f)


def touch(filename: str) -> None:
    with open(filename, "w"):
        pass



import os
import shutil
from typing import List

from pydmt.api.builder import Builder
from pydmt.core.utils import sha1_files_folders, files_under_folder, unlink_files, copy_mkdir
import subprocess


"""
This is review of how to build a sphinx documentation:
- if you want documentation for the code you need to run "sphinx-apidoc"
- it will generated 
"""


class Sphinx(Builder):
    def get_sources(self) -> List[str]:
        return self._get_source_folder_real()

    def __init__(self,
                 package_name: str,
                 source_folder: str = "sphinx",
                 target_folder: str = "docs"):
        super().__init__()
        self.package_name = package_name
        self.source_folder = source_folder
        self.target_folder = target_folder

    def get_signature(self) -> str:
        if os.path.isfile(self.package_name+".py"):
            files = [self.package_name+".py"]
            files.extend(self._get_source_folder_real())
            return sha1_files_folders(
                files=files,
                folders=[
                    os.path.join(self.source_folder, "static"),
                    os.path.join(self.source_folder, "copy"),
                ],
            )
        if os.path.isdir(self.package_name):
            return sha1_files_folders(
                files=self._get_source_folder_real(),
                folders=[
                    os.path.join(self.source_folder, "static"),
                    os.path.join(self.source_folder, "copy"),
                    self.package_name,
                ]
            )
        raise ValueError("Sphinx cannot find source code")

    def _get_source_folder_targets(self) -> List[str]:
        return [
            os.path.join(self.source_folder, "modules.rst"),
            os.path.join(self.source_folder, "{}.rst".format(self.package_name)),
            # We need to add the list of all output files of running sphinx-apidoc
            # os.path.join(self.source_folder, "{}.scripts.rst".format(self.package_name)),
        ]

    def _get_source_folder_real(self) -> List[str]:
        return [
            os.path.join(self.source_folder, "index.rst"),
            os.path.join(self.source_folder, "conf.py"),
        ]

    def build(self) -> None:
        unlink_files(self._get_source_folder_targets(), only_if_exist=True)
        args = [
            "sphinx-apidoc",
            "-f",
            "-o",
            self.source_folder,
        ]
        if os.path.isfile(self.package_name+'.py'):
            args.append(self.package_name+'.py')
        else:
            args.append(self.package_name)
        subprocess.check_call(args)
        if os.path.isdir(self.target_folder):
            shutil.rmtree(self.target_folder, ignore_errors=False)
            subprocess.check_call([
                "sphinx-build",
                # don't use a saved environment, always read all files
                # "-E",
                # Do not emit colored output(default: auto - detect)
                "--no-color",
                # turn warnings into errors
                "-W",
                # no output on stdout, just warnings on stderr
                "-q",
                self.source_folder,
                self.target_folder,
            ])
        for filename in files_under_folder(os.path.join(self.source_folder, "copy")):
            basename = os.path.basename(filename)
            copy_mkdir(filename, os.path.join(self.target_folder, basename))

    def get_targets_post_build(self) -> List[str]:
        return_list = self._get_source_folder_targets()
        return_list.extend(files_under_folder(self.target_folder))
        return return_list



from typing import List

from pydmt.api.builder import Builder
from pydmt.core.utils import sha1_file


class Fail(Builder):
    def get_sources(self) -> List[str]:
        return [self.source]

    def __init__(self, source: str, target: str):
        super().__init__()
        self.source = source
        self.target = target

    def get_signature(self) -> str:
        return sha1_file(self.source)

    def build(self):
        raise ValueError("fail")

    def get_targets(self) -> List[str]:
        return [self.target]



import os
import pylogconf.core
import sys

from pydmt.builders.sphinx import Sphinx
from pydmt.core.pydmt import PyDMT
from pydmt.features.templating import Templating


def main():
    debug = False
    sphinx = True
    add_import_of_cwd = True
    if add_import_of_cwd:
        if sys.path[0] != "":
            sys.path.insert(0, "")
    pylogconf.core.setup()
    if debug:
        print("system path is [{}]".format(sys.path))
    p = PyDMT()
    t = Templating()
    t.setup(p)
    if sphinx and os.path.isdir("sphinx"):
        import config.project
        p.add_builder(Sphinx(package_name=config.project.project_name))
    p.build_all()


if __name__ == '__main__':
    main()



import abc
from typing import List


class Builder(metaclass=abc.ABCMeta):
    """
    A Builder is what really builds things in the system.
    A builder knows on which inputs it relies and what
    outputs it generates (sometimes before build and sometimes
    after).
    """
    __metaclass__ = abc.ABCMeta

    def get_name(self) -> str:
        return self.__class__.__name__+" "+",".join(self.get_sources())+"->"+",".join(self.get_targets())

    @abc.abstractmethod
    def __init__(self):
        """ initialize your builder here """
        pass

    @abc.abstractmethod
    def build(self):
        """ this method actually does the building """
        pass

    @abc.abstractmethod
    def get_sources(self) -> List[str]:
        return []

    def get_targets(self) -> List[str]:
        """ return the names of the files that you are sure you produce. If you do not know them,
        you must implement the get_results_names_post_build method and return None here."""
        return []

    def get_targets_post_build(self) -> List[str]:
        """ find out the names of the files that you produced.
        If you implemented the get_results_names and there you return all the results
        that you created then you don't need to implement anything here """
        return []

    @abc.abstractmethod
    def get_signature(self) -> str:
        """ return the sha1 of anything that identifies the build """
        pass
